/***************************************************************************
	Routines for supporting the launch of a GAMMA job on the cluster.
	Read COPYRIGHT file in this directory
****************************************************************************/

/*
 *  Try to create a virtual GAMMA across the network, assigning it a unique
 *  par_pid.
 *  On success it returns the par_pid, and a list of indexes of the nodes
 *  involved in the virtual GAMMA as its 2th arg.
 *  NOTE:  the arguments "user_space_node_list_***" point to user space!!!
 */
long
sys_gamma_vm_create
(
	unsigned long num_instances,
	void *user_space_node_list_in,
	void *user_space_node_list_out
)
{

	int found, j, procnum;
	unsigned long i,par_pid;
	par_entry *p;
	port_entry *pp;
	time_586 a;


	if (num_instances < 1
		 || num_instances > total_num_nodes*MAX_NUM_LOCAL_INST) {
printg("sys_gamma_vm_create(): wrong or unfeasible no. of instances\n");
		return -GAMMA_ERR_GENERIC;
	}

/*  choose a par_pid, if one is available  */
/*  par_pid #0 is reserved!  */
	found = 0;
	for (par_pid=1; par_pid <= MAX_PAR_PID; par_pid++)
		if (
			(par_proc_table[par_pid] == NULL)
			&&
			(!((par_pid-1-phys_whoami) % total_num_nodes))
		) {
			found = 1;
			break;
		}
	if (!found) {
printg("sys_gamma_vm_create(): could not find a par_pid\n");
		return -GAMMA_ERR_GENERIC;
	}

/*  Allocate a descriptor for the new virtual GAMMA  */
	par_proc_table[par_pid] = (par_entry *)
		__get_free_pages(GFP_KERNEL,par_entry_magnitude);
	p = par_proc_table[par_pid];
	if (p == NULL) {
printg("sys_gamma_vm_create(): kmalloc() failed\n");
		return -GAMMA_ERR_GENERIC;
	}

	memset(p,0,sizeof(par_entry));

/*  No process instances yet  */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		p->inst_table[i].pid = -1;
		p->inst_table[i].task = NULL;
		p->inst_table[i].whoami = -1;
	}

	p->vm.num_instances = num_instances;

	procnum = 0;

	if (user_space_node_list_in == NULL) {

/*
 *  Choose "num_instances" nodes to create a virtual GAMMA,
 *  indexing them from 0 on, fill in the virtual GAMMA
 *  descriptor, find out how many process instances are allocated on the
 *  local host ("procnum"), and allocate room for pre-computed pkt headers.
 *  Nodes are chosen circularly in file /etc/gamma.conf, starting
 *  from local host.
 */

		for (i=0; i<num_instances; i++) {
			p->vm.node_list[i] = (phys_whoami+i) % total_num_nodes;
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printg("sys_gamma_vm_create():  too many proc instances on this node.  Aborting...\n");
				par_proc_table[par_pid] = NULL;
				free_pages(
					(unsigned long)p,par_entry_magnitude);
				return -GAMMA_ERR_GENERIC;
			}
			p->inst_table[procnum-1].whoami = i;
			for (j=0; j<=MAX_PORT_NUMBER; j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->task = NULL;
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_intra_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = kmalloc(
						HEADER_SIZE+MAX_1ST_PIECE_SIZE,
						GFP_KERNEL);
				if (pp->header == NULL) {
printg("sys_gamma_vm_create():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
			}
		}

	} else {

/*
 *  Take "num_instances" nodes from the user-provided list of machines,
 *  then build the virtual GAMMA similar to the above "if" branch.
 */
		gamma_copy_from_user(
			p->vm.node_list,user_space_node_list_in,
			num_instances*sizeof(unsigned char)
		);

		for (i=0; i<num_instances; i++) {
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printg("sys_gamma_vm_create():  too many proc instances on this node.  Aborting...\n");
				par_proc_table[par_pid] = NULL;
				free_pages(
					(unsigned long)p,par_entry_magnitude);
				return -GAMMA_ERR_GENERIC;
			}
			p->inst_table[procnum-1].whoami = i;
			for (j=0; j<=MAX_PORT_NUMBER; j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->task = NULL;
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_intra_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = kmalloc(
						HEADER_SIZE+MAX_1ST_PIECE_SIZE,
						GFP_KERNEL);
				if (pp->header == NULL) {
printg("sys_gamma_vm_create():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
			}
		}

	}

	p->num_locinst = procnum;

/*  Return the chosen nodes to the caller in user space  */
	gamma_copy_to_user(
		user_space_node_list_out,p->vm.node_list,
		num_instances*sizeof(unsigned char)
	);

	gamma_time(a);
	p->vm.random_key = a.lo;

/*  Register the virtual GAMMA across the network, via broadcast  */
/*  This is repeated, because some switches tend to "sleep" sometimes */
	for (j=3;j;j--)
		gamma_kernel_send(
			BROADCAST,par_pid,VM_CREATE,
			(void *)(&(p->vm)),sizeof(vm_descriptor),1
		);

	return par_pid;

}


long
sys_gamma_vm_create_pid
(
 	unsigned long par_pid,
	unsigned long num_instances,
	void *user_space_node_list
)
{

	unsigned long i;
	int j,procnum;
	par_entry *p;
	port_entry *pp;
	time_586 a;


	if (num_instances < 1
		|| num_instances > total_num_nodes*MAX_NUM_LOCAL_INST) {
printg("sys_gamma_vm_create_pid(): wrong or unfeasible no. of instances\n");
		return -GAMMA_ERR_GENERIC;
	}

/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	p = par_proc_table[par_pid];

/*  A virtual GAMMA should already exist!  */
	if (p == NULL) {
printg("sys_gamma_vm_create_pid(): par_pid not found\n");
		return -GAMMA_ERR_GENERIC;
	}

/*  We should have collected "num_instances" participants already!  */
	if (p->vm.num_instances != num_instances) {
printg("sys_gamma_vm_create_pid(): num_instances mismatch\n");
		return -GAMMA_ERR_GENERIC;
	}

/*
 *  Collect participant nodes, init the virtual GAMMA, give instance
 *  number 0 to local proc (master).
 */

	p->inst_table[0].whoami = 0;

	for (i=0; i<num_instances; i++) {
		for (j=0; j<MAX_NUM_NODES*MAX_NUM_LOCAL_INST; j++)
			if (p->wish_flag[j] > 0) {
				p->wish_flag[j]--;
				break;
			}
			else
/*  "wish_flag" is negative if relative to the node of master proc  */
			if (p->wish_flag[j] < 0) {
				p->wish_flag[j]++;
				break;
			}
		if (j>=MAX_NUM_NODES*MAX_NUM_LOCAL_INST) {
			/*  Should never arrive here!  */
printg("sys_gamma_vm_create_pid(): should not have arrived here\n");
			return -GAMMA_ERR_GENERIC;
		}
		p->vm.node_list[i] = j;
	}

	procnum = 0;
	for (i=0; i<num_instances; i++) {
		if (p->vm.node_list[i] == phys_whoami) {
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printg("sys_gamma_vm_create_pid():  too many proc instances on this node.  Aborting...\n");
				par_proc_table[par_pid] = NULL;
				free_pages(
					(unsigned long)p,par_entry_magnitude);
				return -GAMMA_ERR_GENERIC;
			}
			p->inst_table[procnum-1].whoami = i;
			for (j=0; j<=MAX_PORT_NUMBER; j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->task = NULL;
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_intra_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = kmalloc(
						HEADER_SIZE+MAX_1ST_PIECE_SIZE,
						GFP_KERNEL);
				if (pp->header == NULL) {
printg("sys_gamma_vm_create_pid():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
		}
	}

	p->num_locinst = procnum;

/*  Return the chosen nodes to the caller in user space.  */
	gamma_copy_to_user(
		user_space_node_list,p->vm.node_list,
		num_instances*sizeof(unsigned char)
	);

	gamma_time(a);
	p->vm.random_key = a.lo;

/*  Register the virtual GAMMA across the network, via broadcast  */
/*  This is repeated, because some switches tend to "sleep" sometimes */
	for (j=3;j;j--)
		gamma_kernel_send(
			BROADCAST,par_pid,VM_CREATE,
			(void *)(&(p->vm)),sizeof(vm_descriptor),1
		);

	return par_pid;

}


/*
 *  The invoking process informs all nodes that it would need a
 *  virtual GAMMA identified by "par_pid" and expects "num_instances"
 *  other processes to join the virtual GAMMA as well.
 */
long
sys_gamma_vm_wish
(unsigned long par_pid, unsigned long num_instances, int first_time)
{

	unsigned long i;
	par_entry *p;
	time_586 a;


	if (num_instances < 1
		|| num_instances > total_num_nodes*MAX_NUM_LOCAL_INST)
		return -GAMMA_ERR_GENERIC;

/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

/*
 * this can be necessary to let more GAMMA jobs start simultaneously on the
 * same cluster
 */
/* NOTE: "0" is purely arbitrary here, and possibly wrong, but harmless */
	sys_gamma_poll(0);

	if (first_time)
	if (par_proc_table[par_pid] == NULL) {
/*  Allocate a descriptor for the new virtual GAMMA  */
		par_proc_table[par_pid] = (par_entry *)
			__get_free_pages(GFP_KERNEL,par_entry_magnitude);
		p = par_proc_table[par_pid];
		if (p == NULL) {
printg("sys_gamma_vm_wish(): kmalloc() failed\n");
			return -GAMMA_ERR_GENERIC;
		}
		memset(p,0,sizeof(par_entry));
		for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
			p->inst_table[i].pid = -1;
			p->inst_table[i].task = NULL;
			p->inst_table[i].whoami = -1;
		}
		for (i=0; i<MAX_NUM_NODES*MAX_NUM_LOCAL_INST; i++)
			p->wish_flag[i] = 0;
	}

	p = par_proc_table[par_pid];
	if (p == NULL)
		return -GAMMA_ERR_GENERIC;

	if (p->vm.num_instances < num_instances) {
		if (first_time) {
/*
 *  Local proc expresses its wish to join the
 *  virtual GAMMA #par_pid, and tells that to
 *  every node in the cluster.
 */
			p->vm.num_instances++;
			p->wish_flag[phys_whoami]++;
			gamma_kernel_send(BROADCAST,par_pid,VM_WISH,NULL,0,1);
		}
		return -GAMMA_ERR_TRYAGAIN;
	}

/*  All procs have expressed their wish to join the virtual GAMMA  */

/*
 *  Is the local proc master or slave?  The rule is:
 *  The process allocated to the node with lowest physical rank
 *  will be the master.  If more processes are on that node, follow the FCFS
 *  policy (mutual exclusion is required by this routine).
 */
	for (i=0; i<MAX_NUM_NODES*MAX_NUM_LOCAL_INST; i++) {
		if (!p->wish_flag[i])
			continue;
		if (p->wish_flag[i] < 0)
/*
 *  Master is another local proc instance (why?  See below...)
 *  The caller will be slave.
 */
			return 1;
		if (i == phys_whoami) {
/*
 *  The caller will be master.  Changing the sign to "wish_flag" prevents other
 *  local proc instances to take the same decision.
 */
			p->wish_flag[i] *= -1;

			gamma_time(a);
			p->vm.random_key = a.lo;

			return 0;
		}
/*  The caller will be slave   */
		return 1;
	}

/*  The caller will be slave   */
	return 1;

}


/*
 *  The invoking process needs the mapping "instance # --> physical node #" .
 *  This functionality is also contained in sys_gamma_vm_create() and
 *  sys_gamma_vm_create_pid().
 *  NOTE:  the argument "user_space_node_list" points to user space!!!
 */
long
sys_gamma_vm_getinfo
(unsigned long par_pid, void *user_space_node_list)
{

/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	if (par_proc_table[par_pid] == NULL)
		return -GAMMA_ERR_GENERIC;

	gamma_copy_to_user(
	  user_space_node_list,par_proc_table[par_pid]->vm.node_list,
	  par_proc_table[par_pid]->vm.num_instances*sizeof(unsigned char)
	);

	return 0;

}


/*
 * Read in the config file, possibly clean some data struct
 * returns the physical rank of this machine in the cluster.
 */
long
sys_gamma_getconfig
(
	unsigned long n,
	void *user_space_eth_table,
	void *user_space_routing_table, /* != NULL iff #defined MORE_NICS... */
	unsigned long flag
)
{

	unsigned long i,j,k,n_old;
	long res;


	if (flag) {

/*  Init and clean a lot of thing ;-)  */

		if (n < 1 || n > MAX_NUM_NODES) {
printg("sys_gamma_getconfig(): wrong or unfeasible number of nodes\n");
			return -GAMMA_ERR_GENERIC;
		}

		if (total_num_nodes)
			n_old = total_num_nodes;
		else
/*
 *  This might be first invocation after boot (that is, after gamma_start()).
 *  Thus, we have allocated MAX_NUM_NODES Tx tmp rings (presumably more than
 *  needed).
 */
			n_old = MAX_NUM_NODES;
/*  De-allocate Tx tmp rings.  */
		for (i=0; i<n_old; i++)
			for (j=0; j<GAMMA_TX_WIN_SIZE; j++)
				if (tx_tmp_ring[i].buf[j].b != NULL)
					kfree(tx_tmp_ring[i].buf[j].b);

		gamma_initialized = 1;
		total_num_nodes = n;

/*
 *  Set the initial credit towards each destination.
 *  It may be safer to leave two (or more?) available Rx slots
 *  off-credit for service (non-user) communications.
 */
		for (i=0; i<total_num_nodes; i++) {
			initial_credit[i] = credit_size;
			credit_ceil[i] = credit_size;
			credit_growth_cnt[i] = 0;
			atomic_set(&local_credit[i],initial_credit[i]);
		}

/*  Allocate and clean the Tx tmp rings.  */

		for (i=0; i<total_num_nodes; i++) {
			for (j=0; j<GAMMA_TX_WIN_SIZE; j++) {
				tx_tmp_ring[i].buf[j].b = (unsigned char *)
					kmalloc(FRAME_SIZ, GFP_KERNEL);
				if (tx_tmp_ring[i].buf[j].b == NULL) {
printg("sys_gamma_getconfig():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
				tx_tmp_ring[i].buf[j].l = 0;
			}
			tx_tmp_ring[i].tail = 0;
#ifdef SEQCHECK
/*  Reset the outgoing pkt numbering.  */
			tx_tmp_ring[i].tx_seq = 0;
#endif
		}
		for (j=0; j<GAMMA_TX_WIN_SIZE; j++)
			kernel_tx_tmp_ring.buf[j].l = 0;
		kernel_tx_tmp_ring.tail = 0;
#ifdef SEQCHECK
/*  Reset the outgoing pkt numbering.  */
		kernel_tx_tmp_ring.tx_seq = 0;

/*  Reset the tables of incoming pkt numbering and congestion flags.  */
		for (i=0; i<total_num_nodes; i++) {
			rx_seq[i] = rx_seq_bcast[i] = 0;
			rx_discard[i] = 0;
			congested[i] = 0;
		}
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
		for (j=0; j<MAX_NUM_LOCAL_INST; j++) {
			shared_ring[j].head = 0;
			atomic_set(&shared_ring[j].dirty, 0);
			atomic_set(&shared_ring[j].tail, 0);
			for (i=0; i<SHARED_RING_SIZE; i++) {
				shared_ring[j].desc[i].par_pid = 0;
				shared_ring[j].desc[i].procnum = 0;
				shared_ring[j].desc[i].port = 0;
				shared_ring[j].desc[i].cnk_len = 0;
				shared_ring[j].desc[i].bitmap = IS_EMPTY;
				shared_ring[j].desc[i].watch_time.hi = 0;
				shared_ring[j].desc[i].watch_time.lo = 0;
			}
		}
#endif
		gamma_must_yield.task = NULL;
		gamma_must_yield.time.hi = 0;
		gamma_must_yield.time.lo = 0;
		gamma_run_handler.task = NULL;
#ifdef MORE_INSTANCES_ON_SAME_NODE
		gamma_must_yield_intranode.task = NULL;
		gamma_must_yield.time.hi = 0;
		gamma_must_yield.time.lo = 0;
		gamma_run_handler_intranode = NULL;
#endif

/*  get the table of MAC addresses from user space  */
		memset(eth_table,0,MAX_NUM_NODES*sizeof(eth_addr));
		gamma_copy_from_user(
			eth_table,user_space_eth_table,
			total_num_nodes*sizeof(eth_addr)
		);

#ifdef MORE_NICS_SAMENODE_ROUTING
/* copy the routing table from user space */
		gamma_copy_from_user(
			routing_table,user_space_routing_table,
			MAX_NUM_NODES*sizeof(unsigned short)
		);
#endif

	}

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG(
	printg("sys_gamma_getconfig():  ROUTING_TABLE:\n");
	for (i=0;i<total_num_nodes;i++)
		printg("MAC: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x nic: %d\n",
		       eth_table[i].addr[0],
		       eth_table[i].addr[1],
		       eth_table[i].addr[2],
		       eth_table[i].addr[3],
		       eth_table[i].addr[4],
		       eth_table[i].addr[5],
		       routing_table[i]
		);
)
#else
DEBUG(
	printg("sys_gamma_getconfig():  ETH_TABLE:\n");
	for (i=0; i<total_num_nodes; i++)
               printg("%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		      eth_table[i].addr[0],
		      eth_table[i].addr[1],
		      eth_table[i].addr[2],
		      eth_table[i].addr[3],
		      eth_table[i].addr[4],
		      eth_table[i].addr[5]
	       );
)
#endif

/*  what node is this wrt the physical GAMMA?  */
	res = -1;
	for (i=0;i<total_num_nodes;i++) {
		for(k=0; k<num_gamma_devices; k++) {
			for (j=0;j<6;j++)
				if (eth_table[i].addr[j] != 
				    gamma_devices[k]->dev_addr[j])
					break;
			if (j==6) {
				res = i;
				goto end;
			}
		}
	}

 end:

	if (res < 0) {
printg("sys_gamma_getconfig(): local MAC addr not found in gamma.conf\n");
		return -GAMMA_ERR_GENERIC;
	}

DEBUG( printg("sys_gamma_getconfig(): phys_whoami = %ld\n", res) );
DEBUG( printg("sys_gamma_getconfig(): gamma_physical_nodes = %ld\n",
total_num_nodes) );

	phys_whoami = res;
	return res;

}


/*
 *  Registers the PID and the task pointer of the calling process
 *  with the virtual GAMMA specified by par_pid (which is supposed to exist).
 */
long
sys_gamma_register
(unsigned long par_pid)
{

	unsigned long i;
	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	p = par_proc_table[par_pid];
	for (i=100;i;i--) {
		if (p) break;
/* NOTE: "0" is purely arbitrary here, and possibly wrong, but harmless */
		sys_gamma_poll(0);  /* gamma_kernel_rec() eventll VM_CREATE */
		busywait(1000);
		p = par_proc_table[par_pid];
	}

	if (!i && p == NULL) {
/*  not an existing virtual GAMMA  */

printg("sys_gamma_register(): par_pid %ld not created\n",par_pid);

		return -GAMMA_ERR_GENERIC;
	}

DEBUG(
printg("sys_gamma_register(): virtual GAMMA %ld:\nnum_instances %u\n",
	par_pid,
	p->vm.num_instances);
for (i=0; i<p->vm.num_instances; i++)
	printg("%lu: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		i,
		eth_table[p->vm.node_list[i]].addr[0],
		eth_table[p->vm.node_list[i]].addr[1],
		eth_table[p->vm.node_list[i]].addr[2],
		eth_table[p->vm.node_list[i]].addr[3],
		eth_table[p->vm.node_list[i]].addr[4],
		eth_table[p->vm.node_list[i]].addr[5]
	);
);

/*
 *  Since more process instances are allowed on the same host, the
 *  caller has to find the right slot in the virtual GAMMA.
 *  Mutex is mandatory; re-use gamma_tx_lock for this.
 */
	spin_lock(&gamma_tx_lock);
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		if (p->inst_table[i].pid != -1)
/*  Another process already registered in this slot  */
			continue;
		if (p->inst_table[i].whoami >= 0) {
/*  A free, valid slot was found  */
			p->inst_table[i].pid = current->pid;
			p->inst_table[i].task = current;

DEBUG_L(printg(
"sys_gamma_register(): virtual GAMMA %ld: whoami %d (pid %d)\n",
par_pid,p->inst_table[i].whoami,p->inst_table[i].pid
));

			spin_unlock(&gamma_tx_lock);
			return 0;
		}

DEBUG(printg(
"sys_gamma_register(): local host does not belong to virtual GAMMA %ld\n",
par_pid
));

		spin_unlock(&gamma_tx_lock);
		return -GAMMA_ERR_GENERIC;
	}

printg("sys_gamma_register(): par_pid %ld: could not register\n",par_pid);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

}


long
sys_gamma_register_nospawn
(unsigned long par_pid)
{

	unsigned long i;
	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

/*
 * this can be necessary to let more GAMMA jobs start simultaneously on the
 * same cluster
 */
/* NOTE: "0" is purely arbitrary here, and possibly wrong, but harmless */
	sys_gamma_poll(0);

	p = par_proc_table[par_pid];

	if (p == NULL) {
/*  not an existing virtual GAMMA  */
printg("sys_gamma_register_nospawn(): par_pid %ld not created\n",par_pid);
		return -GAMMA_ERR_GENERIC;
	}

DEBUG(
{
int i;
printg("sys_gamma_register_nospawn():  par_pid %ld, pid %d\n",
	par_pid,
	p->inst_table[0].pid
);
printg("Virtual GAMMA %ld:\nnum_instances %d\n",
	par_pid,
	p->vm.num_instances);
for (i=0; i<p->vm.num_instances; i++)
	printg("%d: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		i,
		eth_table[p->vm.node_list[i]].addr[0],
		eth_table[p->vm.node_list[i]].addr[1],
		eth_table[p->vm.node_list[i]].addr[2],
		eth_table[p->vm.node_list[i]].addr[3],
		eth_table[p->vm.node_list[i]].addr[4],
		eth_table[p->vm.node_list[i]].addr[5]
	);
}
);

/*
 *  Since more process instances are allowed on the same station, the
 *  caller has to find the right slot in the virtual GAMMA.
 */
	spin_lock(&gamma_tx_lock);
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		if (p->inst_table[i].pid != -1)
/*  Another process already registered in this slot  */
			continue;
		if (p->inst_table[i].whoami >= 0) {
/*  A free, valid slot was found  */
DEBUG(printg(
"sys_gamma_register_nospawn(): procnum %ld, whoami %d\n",
i,p->inst_table[i].whoami
));
			p->inst_table[i].pid = current->pid;
			p->inst_table[i].task = current;
			spin_unlock(&gamma_tx_lock);
			return 0;
		}
		else {
/*
 *  The "whoami" field will eventually be filled up by the master process
 *  when calling sys_gamma_vm_create_pid().  This process will try again
 *  until success.
 */
DEBUG(printg(
"sys_gamma_register_nospawn(): local host does not belong to par_pid %ld yet\n",
par_pid
));
			spin_unlock(&gamma_tx_lock);
			return -GAMMA_ERR_TRYAGAIN;
		}
	}
printg("sys_gamma_register_nospawn(): par_pid %ld:  could not register\n",
par_pid);
	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

}


/*
 *  sys_gamma_init, parts 1 & 2 & 3.
 *  These GAMMA primitives look into par_proc_table for the caller's entry
 *  (which was assigned to the caller via a previous invocation to
 *  "gamma_register()")
 *  and then register the caller's tables of receiver buffers and semaphores.
 */

/*
 *  This routine returns the no. of nodes of the virtual GAMMA
 *  which the caller had previously registered to, and
 *  the instance no. of the caller.
 */
long
sys_gamma_init_part1
(
	void *user_space_msglen_ptr
)
{

	int found=0,i=0,par_pid;
	par_entry *p;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		if (par_proc_table[par_pid] == NULL)
			continue;
		for (i=0; i<MAX_NUM_LOCAL_INST; i++)
			if (
				par_proc_table[par_pid]->inst_table[i].pid
				== current->pid
			) {
				found = 1;
				break;
			}
		if (found)
			break;
	}
	if (found) {
		p = par_proc_table[par_pid];
		p->inst_table[i].msglen_ptr =
			user_space_msglen_ptr;
	        return (long)(
			((unsigned long)(p->vm.num_instances) << 16) +
			(unsigned long)p->inst_table[i].whoami
		);
	}
	return -GAMMA_ERR_GENERIC;

}


/*
 *  This routine returns the "procnum" of the caller and the par_pid
 *  of the virtual GAMMA which the caller had previously registered to.
 */
long
sys_gamma_init_part2
(
	void *user_space_sem1_table_ptr,
	void *user_space_pre_sem_table_ptr
)
{

	int found=0,i=0,par_pid;
	par_entry *p;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		if (par_proc_table[par_pid] == NULL)
			continue;
		for (i=0; i<MAX_NUM_LOCAL_INST; i++)
			if (
				par_proc_table[par_pid]->inst_table[i].pid
				== current->pid
			) {
				found = 1;
				break;
			}
		if (found)
			break;
	}
	if (found) {
		p = par_proc_table[par_pid];
/*  Register the tables of user-space semaphores.  */
#ifdef USE_DAMNICK
		p->inst_table[i].sem1_table =
			user_space_sem1_table_ptr;
		p->inst_table[i].pre_sem_table =
			user_space_pre_sem_table_ptr;
#endif
/*  Return the "procnum" of the caller.  */
	        return (long)(
			(unsigned long)(par_pid << 16) +
			(unsigned long)i
		);
	}
	return -GAMMA_ERR_GENERIC;

}


/*
 *  This routine returns the clock freq. in Mhz of the local host.
 */
long
sys_gamma_init_part3
(void)
{

	int found=0,i=0,par_pid;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		if (par_proc_table[par_pid] == NULL)
			continue;
		for (i=0; i<MAX_NUM_LOCAL_INST; i++)
			if (par_proc_table[par_pid]->inst_table[i].pid
				== current->pid
			) {
				found = 1;
				break;
			}
		if (found) break;
	}
	if (found)
/*  Return the CPU clock freq. in MHz  */
	        return cpu_khz/1000;
	return -GAMMA_ERR_GENERIC;

}


long
sys_gamma_exit
(unsigned long par_pid, unsigned long procnum)
{

	par_entry *p;
	int i,found;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int j;
#endif


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	if (!spin_trylock(&gamma_rx_lock))
		return -GAMMA_ERR_TRYAGAIN;
	if (!write_trylock(&gamma_rxintranode_lock)) {
		spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}
	if (!spin_trylock(&gamma_tx_lock)) {
		write_unlock(&gamma_rxintranode_lock);
		spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}

	p = par_proc_table[par_pid];

	if (p == NULL) {
/*  Virtual GAMMA #par_pid does not exist  */
		spin_unlock(&gamma_tx_lock);
		write_unlock(&gamma_rxintranode_lock);
		spin_unlock(&gamma_rx_lock);
		return -GAMMA_ERR_GENERIC;
	}

	if (p->inst_table[procnum].pid != current->pid) {
		spin_unlock(&gamma_tx_lock);
		write_unlock(&gamma_rxintranode_lock);
		spin_unlock(&gamma_rx_lock);
		return -GAMMA_ERR_GENERIC;
	}

/*  unsubscribe from virtual GAMMA  */
	p->inst_table[procnum].pid = -1;
	for (i=0; i<=MAX_PORT_NUMBER; i++)
		if (p->inst_table[procnum].port_table[i].header)
		kfree(p->inst_table[procnum].port_table[i].header);

/*  if no other process instances left, destroy the virtual GAMMA  */
	found = 0;
	for (i=0; i<MAX_NUM_LOCAL_INST; i++)
		if (p->inst_table[i].pid > 0)
			found = 1;
	if (!found) {
		par_proc_table[par_pid] = NULL;
		free_pages((unsigned long)p,par_entry_magnitude);
		gamma_must_yield.task = NULL;
		gamma_must_yield.time.hi = 0;
		gamma_must_yield.time.lo = 0;
		gamma_run_handler.task = NULL;
#ifdef MORE_INSTANCES_ON_SAME_NODE
		gamma_must_yield_intranode.task = NULL;
		gamma_must_yield.time.hi = 0;
		gamma_must_yield.time.lo = 0;
		gamma_run_handler_intranode = NULL;
/*  clean up shared rings  */
		for (j=0; j<MAX_NUM_LOCAL_INST; j++)
			for (i=0; i<SHARED_RING_SIZE; i++)
				if (shared_ring[j].desc[i].par_pid == par_pid)
/* par_pid #0 is reserved, rx_gamma_intranode() will eventually discard */
					shared_ring[j].desc[i].par_pid = 0;
#endif
	}

	spin_unlock(&gamma_tx_lock);
	write_unlock(&gamma_rxintranode_lock);
	spin_unlock(&gamma_rx_lock);

	return 0;

}


/*
 *  Destroy a virtual GAMMA (possibly across the network, flag "global"),
 *  and force all process instances to exit.
 *  It may be invoked by any process instance (usually upon ctl-C or other
 *  dramatic signals)
 *  Flag "rxlock" is 0 if caller is gamma_kernel_rec() (so gamma_rx_lock is
 *  already taken), 1 otherwise (gamma_rx_lock still has to be taken).
 */
long
__sys_gamma_vm_abort
(unsigned long par_pid, int global, int rxlock)
{

	int i,j,found;
	int was_not_already_quitting = 0;
	par_entry *p;


	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	if (rxlock)
		/* gamma_rx_lock not already locked, so take it as well */
		if (!spin_trylock(&gamma_rx_lock))
			return -GAMMA_ERR_TRYAGAIN;
	if (!write_trylock(&gamma_rxintranode_lock)) {
		if (rxlock) spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}
	if (!spin_trylock(&gamma_tx_lock)) {
		write_unlock(&gamma_rxintranode_lock);
		if (rxlock) spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}

	p = par_proc_table[par_pid];

	if (p == NULL) {
/*  Virtual GAMMA #par_pid does not exist  */
		spin_unlock(&gamma_tx_lock);
		write_unlock(&gamma_rxintranode_lock);
		if (rxlock) spin_unlock(&gamma_rx_lock);
		return -GAMMA_ERR_GENERIC;
	}

DEBUG(printg("sys_gamma_vm_abort: found par_pid %ld...\n",par_pid));

	if (global) {
/*  unsubscribe self from virtual GAMMA and dealloc own ports */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		if (p->inst_table[i].pid != current->pid) continue;
		p->inst_table[i].pid = -1;
		for (j=0; j<=MAX_PORT_NUMBER; j++)
			if (p->inst_table[i].port_table[j].header)
			kfree(p->inst_table[i].port_table[j].header);
	}
	} else {
/*  dealloc all instances' ports */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++)
		if (p->inst_table[i].whoami >= 0)
			for (j=0; j<=MAX_PORT_NUMBER; j++)
				if (p->inst_table[i].port_table[j].header)
				kfree(p->inst_table[i].port_table[j].header);

	}

	if (global) {
/*
 *  Destroy the virtual GAMMA across the network and
 *  require remote process instances to quit (if not already done)
 */
		if (!p->quitting) {
			p->quitting = 1;
			was_not_already_quitting = 1;
			for (j=3;j;j--)
				gamma_kernel_send(BROADCAST,par_pid,VM_ABORT,
					&p->vm.random_key,
					sizeof(p->vm.random_key),0);
		}
	}
/*  Quit local process instances, if any  */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++)
		if (
			p->inst_table[i].whoami >= 0 &&
			p->inst_table[i].task != NULL &&
			p->inst_table[i].task->state != EXIT_ZOMBIE &&
			p->inst_table[i].task->state != EXIT_DEAD &&
			p->inst_table[i].task->pid == p->inst_table[i].pid &&
			find_task_by_pid(p->inst_table[i].pid) != NULL
		) {
			if (global) {
				if (was_not_already_quitting)
		 		if (p->inst_table[i].task->pid
					!= current->pid) {

DEBUG_L(printg("sys_gamma_vm_abort: SIGTERM to inst %d\n",i));

					force_sig(SIGTERM,
						p->inst_table[i].task);
				}
			}
			else {

DEBUG_L(printg("sys_gamma_vm_abort: SIGKILL to inst %d\n",i));

				force_sig(SIGKILL, p->inst_table[i].task);
			}
		}

	gamma_must_yield.task = NULL;
	gamma_must_yield.time.hi = 0;
	gamma_must_yield.time.lo = 0;
	gamma_run_handler.task = NULL;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	gamma_must_yield_intranode.task = NULL;
	gamma_must_yield.time.hi = 0;
	gamma_must_yield.time.lo = 0;
	gamma_run_handler_intranode = NULL;
/*  clean up shared rings  */
	for (j=0; j<MAX_NUM_LOCAL_INST; j++)
		for (i=0; i<SHARED_RING_SIZE; i++)
			if (shared_ring[j].desc[i].par_pid == par_pid)
/* par_pid #0 is reserved, rx_gamma_intranode() will eventually discard */
				shared_ring[j].desc[i].par_pid = 0;
#endif

/*  destroy the virtual GAMMA if no other instances still subscribed  */
	found = 0;
	if (global)
		for (i=0; i<MAX_NUM_LOCAL_INST; i++)
			if (p->inst_table[i].pid > 0)
				found = 1;
	if (!found) {
		par_proc_table[par_pid] = NULL;
		free_pages((unsigned long)p,par_entry_magnitude);
	}

DEBUG_L(printg("sys_gamma_vm_abort: par_pid %ld aborted\n",par_pid));


	spin_unlock(&gamma_tx_lock);
	write_unlock(&gamma_rxintranode_lock);
	if (rxlock) spin_unlock(&gamma_rx_lock);

	return 0;

}


long
sys_gamma_vm_abort
(unsigned long par_pid, int global)
{
	return __sys_gamma_vm_abort(par_pid,global,1);
}
