#include "../gamma_def.h"
#include "../lib/libgamma.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include <time.h>
#include <errno.h>

#define NEW_GAMMA_CONF

#define MAX_LEN 64


/*************************
input syntax (default file /etc/gamma.global)

for flat configuration:
--------------------
net {
  node1
  node2
  ....
  nodeN
}
--------------------

for fnn configuration:
--------------------
subnet {
  node1  nic_index
  node2  nic_index
  ....   ...
  nodeN  nic_index
}

.....
.....
.....

subnet {
  node1  nic_index
  node2  nic_index
  ....   ...
  nodeN  nic_index
}

-------------------

*************************/



/*******************
 * DATA STRUCTURES *
 *******************/

/* command line options */
char conf_eth_offset = -1;
char conf_no_eth_res = 0;
char *conf_program_name = NULL;
char conf_try = 0;
char *conf_file_input = "/etc/gamma.global" ;
char *conf_file_output = "/etc/gamma.conf";
char conf_verbose = 0;
char conf_very_verbose = 0;
char *conf_hostname = NULL;
char conf_random = 0;

/* nodes that make up the cluster */
int num_nodes;
char host_table[MAX_NUM_NODES][MAX_LEN];

/* possible multiple routes */
struct route {
  int local;
  int remote;
};
struct route routes[MAX_NUM_NODES][MAX_NUM_NICS];
int routes_num[MAX_NUM_NODES];

/* chosen routes */
eth_addr eth[MAX_NUM_NODES];
unsigned short nic[MAX_NUM_NODES];

/* node we build configuration for */
int my_index;
int my_num_nics;
int my_nics[MAX_NUM_NICS];
int my_first_nic_index;

/* Subnets that build up the fnn network.
 * A subnet is a collection of associations "hostname" - "nic used 
 * on that subnet by that host". */
struct subnet {
  char entry_name[MAX_NUM_NODES][MAX_LEN];
  int entry_nic[MAX_NUM_NODES];
  int num_entries;
};

int num_subnets;
struct subnet *subnets;



/**********
 * PARSER *
 **********/

#define NET 1
#define SUBNET 2
#define OPEN 3
#define CLOSE 4
#define HOSTNAME 5
#define INDEX 6

/* the parsed file contained a flat configuration
 * with "net { node1 node2 ... }" syntax */
char parsed_flat_syntax = 0;

/* line which is currently being read */
int line = 1;

/* parser error messages */
char *err_current = "";
char *err_unclosed_net = "Missing \"}\" or invalid hostname";
char *err_missing_index = "Missing nic index after hostname";
char *err_parsing_index = "Problem parsing nic index";
char *err_missing_open = "Missing \"{\"";
char *err_too_many_nets = "Only one net entry is allowed";
char *err_not_net = "network entries must begin with \"net\" or \"subnet\"";

void exit_out_of_bounds()
{
  fprintf(stderr, "Error: out of core.\n");
  exit(10);
}

void print_error(char *s) 
{
  fprintf(stderr, "Error parsing %s at line %d", conf_file_input, line);
  if (s != NULL)
    fprintf(stderr, ":\n  %s\n", s);
  else
    fprintf(stderr, "\n");
}

int parse_next(int level, char *s, FILE *f) 
{
  int i,c,state;
	
  state = 0;  /* 0 == no "#" encountered, otherwise 1 */

scan_line:

  while ((c = fgetc(f)) != EOF && ((!state && !isgraph(c)) || (state && (c != '\n'))));

  if (!state && c == '#') {
    /* commented line starts, discard it */
    state = 1;
    goto scan_line;
  }

  if (state && c != EOF) {
    /* commented line ended, take first char at next line */
    state = 0;
    goto scan_line;
  }

  ungetc(c, f);
	
  for (i=0; isgraph(c = fgetc(f)) && (c != '#'); i++)
    if (i < MAX_LEN-1)
      s[i] = c;

  ungetc(c, f);

  if (i < MAX_LEN-1)
    s[i] = '\0';
  else
    s[MAX_LEN-1] = '\0';

  if (i == 0) 
    return EOF;
  else if (s[0] == '{')
    return OPEN;
  else if (s[0] == '}')
    return CLOSE;
  else if (level == 0 && !strcmp(s, "net"))
    return NET;
  else if (level == 0 && !strcmp(s, "subnet"))
    return SUBNET;
  else if (!strncmp(s, "eth", 3) && isdigit(s[3]))
    return INDEX;
  else if (isalpha(s[0]))
    return HOSTNAME;
  else 
    return EOF;
}

int parse_net(FILE *f, int is_subnet) 
{
  int type;
  char s[MAX_LEN];
  char *err;
  
  subnets[num_subnets].num_entries = 0;

  for (;;) 
    {
      if ((type = parse_next(2, s, f)) == HOSTNAME) 
	{
	  strncpy(subnets[num_subnets].
		  entry_name[subnets[num_subnets].num_entries], s, MAX_LEN);
	}
      else if (type == CLOSE) 
	{
	  if (num_subnets >= MAX_NUM_NODES)
	    exit_out_of_bounds();
	  num_subnets++;
	  return 0;
	}
      else
	{
	  err_current = err_unclosed_net;
	  return -1;
	}

      if (is_subnet) 
	{
	  if (parse_next(2, s, f) == INDEX) 
	    {
	      subnets[num_subnets].entry_nic[subnets[num_subnets].num_entries] =
		strtol(s+3, &err, 10); // saltiamo "eth"
	      if(*err != '\0')
		{
		  err_current = err_parsing_index;
		  return -1;
		}
	    }
	  else
	    {
	      err_current = err_missing_index;
	      return -1;
	    }
	}
      if (subnets[num_subnets].num_entries >= MAX_NUM_NODES)
	exit_out_of_bounds();
      subnets[num_subnets].num_entries++;
    }
}

int parse_file(FILE *f) 
{
  int type;
  char s[MAX_LEN];
  int found_flat_net = 0;
  int found_fnn_net = 0;

  while ((type = parse_next(0, s, f)) != EOF) 
    {
      if (type == NET)
	{
	  if (found_flat_net || found_fnn_net)
	    {
	      err_current = err_too_many_nets;
	      goto bad_exit;
	    }
	  parsed_flat_syntax = found_flat_net = 1;

	  if (parse_next(1, s, f) == OPEN) 
	    {
	      if (parse_net(f, 0) < 0)
		goto bad_exit;
	    }
	  else
	    {
	      err_current = err_missing_open;
	      goto bad_exit;
	    }
	}
      else if (type == SUBNET) 
	{
	  if (found_flat_net)
	    {
	      err_current = err_too_many_nets;
	      goto bad_exit;
	    }
	  parsed_flat_syntax = !(found_fnn_net = 1);

	  if (parse_next(1, s, f) == OPEN) 
	    {
	      if (parse_net(f, 1) < 0) 
		goto bad_exit;
	    }
	  else
	    {
	      err_current = err_missing_open;
	      goto bad_exit;
	    }
	}
      else
	{
	  err_current = err_not_net;
	  goto bad_exit;
	}
    }

  return 0;
      
 bad_exit:
  print_error(err_current);
  return -1;
}



/**********************
 * ROUTES COMPUTATION *
 **********************/

/** search functions **/

int find_in_nodes(char *hostname)
{
  int i;
  
  for (i=0; i < num_nodes; i++) 
    {
      if (!strcmp(hostname, host_table[i]))
	return i;
    }
  return -1;
}

int find_in_subnet(int me, int nic_index, int subnet)
{
  int i;
	
  for (i=0; i < subnets[subnet].num_entries; i++) 
    {
      if (!strcmp(subnets[subnet].entry_name[i], host_table[me]) && 
	  subnets[subnet].entry_nic[i] == nic_index )
	return i;
    }
  return -1;
}


/** compute global node list **/

int compute_fnn_nodes(char *hostname)
{
  int i,j;
  
  for (i=0; i < num_subnets; i++) 
    {
      for (j=0; j < subnets[i].num_entries; j++) 
	{
	  if (!strcmp(subnets[i].entry_name[j], hostname))
	    {
	      if (my_num_nics >= MAX_NUM_NICS)
		exit_out_of_bounds();
	      my_nics[my_num_nics++] = subnets[i].entry_nic[j];
	    }
	  if (find_in_nodes(subnets[i].entry_name[j]) < 0) 
	    {
	      if (num_nodes >= MAX_NUM_NODES)
		exit_out_of_bounds();
	      strcpy(host_table[num_nodes++], subnets[i].entry_name[j]);
	    }
	}
    }
  if ((my_index = find_in_nodes(hostname)) < 0)
    {
      fprintf(stderr, "Error: couldn't find myself (%s) in node list\n", hostname);
      return -1;
    }

  for (i=0, my_first_nic_index=my_nics[0]; i < my_num_nics; i++)
    if (my_first_nic_index > my_nics[i])
      my_first_nic_index = my_nics[i];

  for (i=0; i < num_nodes; i++)
    routes_num[i] = 0;

  return num_nodes;
}

#if 0
void compute_flat_nodes()
{
  int i;

  num_nodes = subnets[0].num_entries;
  for (i=0; i < num_nodes; i++)
    {
      strcpy(host_table[i], subnets[0].entry_name[i]);
      routes_num[i] = 1;
      routes[i][0].remote = conf_eth_offset;
    }
}
#endif


/** functions which compute all possible routes **/

void fill_partial_table(int nic_idx, int subnet, int me_subnet)
{
  int i, entry;
  
  for (i=0; i < subnets[subnet].num_entries; i++) 
    {
      if (i != me_subnet) 
	{
	  entry = find_in_nodes(subnets[subnet].entry_name[i]);
	  routes[entry][routes_num[entry]].local = nic_idx;
	  routes[entry][routes_num[entry]].remote = 
	    subnets[subnet].entry_nic[i];
	  routes_num[entry]++;
	}
    }
}

int fill_tables(char *hostname)
{
  int i,j;
  int me_subnet;
  
  for (i=0; i < my_num_nics; i++) 
    {
      for (j=0; j < num_subnets; j++) 
	{
	  me_subnet = find_in_subnet(my_index , my_nics[i], j);
	  if (me_subnet >= 0) 
	    {
	      fill_partial_table(my_nics[i], j, me_subnet);
	      break;
	    }
	}
    }

  routes[my_index][0].local = routes[my_index][0].remote = my_first_nic_index;
  routes_num[my_index] = 1;
  
  return num_nodes;
}


/** functions which resolve multiple routes **/

int resolve_routes_random(void)
{
  int i, chosen_route, ret = 1;

  for (i=0; i < num_nodes; i++)
    {
      if (routes_num[i] == 0)
	{
	  fprintf(stderr, "Warning: no route to host %s\n", host_table[i]);
	  nic[i] = ret = 0;
	}
      else if (routes_num[i] == 1)
	{
	  nic[i] = routes[i][0].local;
	}
      else if (routes_num[i] > 1)
	{
	  chosen_route = 
	    (int) ( ((float)routes_num[i]) * rand() / (RAND_MAX + 1.0) );
	  routes[i][0].local = routes[i][chosen_route].local;
	  routes[i][0].remote = routes[i][chosen_route].remote;
	  nic[i] = routes[i][0].local;
	}
    }

  return ret;
}

int resolve_routes_static(void)
{
  int i, ret = 1;

  for (i=0; i < num_nodes; i++)
    {
      if (routes_num[i] == 0)
	{
	  fprintf(stderr, "Warning: no route to host %s\n", host_table[i]);
	  nic[i] = ret = 0;
	}
      else
	nic[i] = routes[i][0].local;
    }

  return ret;
}


/** functions for retrieving remote mac addresses and nic indices **/

#define BUFLEN 128

int ask_eth_addr(char *hostname, int nic, eth_addr *e)
{
  int pid, i = 0;
  char *args[8];
  char buffer[BUFLEN];
  char *s, *next;
  int pfds[2];


  pipe(pfds);

  pid = fork();
  
  if(!pid) // child
    {
      close(pfds[0]);
      args[i++] = "/usr/bin/ssh";
      args[i++] = hostname;
      snprintf(buffer, BUFLEN,
      "/sbin/ifconfig -a | grep eth%d | cut -f 11 -d \" \"", nic);
      args[i++] = buffer;
      args[i] = NULL;
      
      close(1);
      dup(pfds[1]);

      execve(args[0], args, NULL);
      perror("Problem with execve()");
    }

  close(pfds[1]);  

  if (read(pfds[0], buffer, BUFLEN) <= 0)
    goto bad_exit;
  
  s = buffer;

  for(i=0; (isalnum(buffer[i]) || buffer[i] == ':') && i < BUFLEN; i++)
    ;
  buffer[i] = '\0';

  if(buffer[0] == '\0')
    goto bad_exit;

  for (i=0; i<6; i++)
    {
      e->addr[i] = strtol(s, &next, 16);
      if(*next == ':')
	s = next+1;
      else break;
    }
  
  e->addrl =
    ((unsigned long)e->addr[0]) +
    ((unsigned long)e->addr[1] << 8) +
    ((unsigned long)e->addr[2] << 16) +
    ((unsigned long)e->addr[3] << 24);
  
  e->addrw =
    ((unsigned long)e->addr[4]) +
    ((unsigned long)e->addr[5] << 8);

  return *next == '\0';

 bad_exit:
  for (i=0; i<6; i++) 
    e->addr[i] = 0;
  e->addrw = 0;
  e->addrl = 0;
  return 0;
}

void resolve_ethernet_addresses(void)
{
  int i, j, nic;
  eth_addr ead;

  for (i=0; i < num_nodes; i++)
    {
      if (conf_no_eth_res || routes_num[i] == 0)
	{
	  for (j=0; j<6; j++) 
	    ead.addr[j] = 0;
	  ead.addrw = 0;
	  ead.addrl = 0;
	}
      else
	{
	  nic = routes[i][0].remote;
	  
	  if (!ask_eth_addr(host_table[i], nic, &ead))
	    fprintf(stderr, "Warning: couldn't obtain ethernet "
		    "address of node %s, nic eth%d.\n", host_table[i], nic);
	}
      eth[i] = ead;
    }
}

#if 0
void get_offset(void)
{
  int pid;
  char buffer[BUFLEN];
  int pfds[2];


  pipe(pfds);

  pid = fork();
  
  if(!pid) // child
    {
      close(pfds[0]);
 
      close(1);
      dup(pfds[1]);

      if (system("/bin/dmesg | grep \"GAMMA  Genoa Active Message MAchine\" | cut -f 10 -d \" \"") != 0)
	{
	  fprintf(stderr,"Problem with system()\n");
	}
      exit(0);
    }

  close(pfds[1]);  

  if (read(pfds[0], buffer, BUFLEN) <= 0)
    goto bad_exit;
  
  if (strncmp(buffer, "eth", 3) || !isdigit(buffer[3]))
    goto bad_exit;
  
  conf_eth_offset = buffer[3] - '0';
  return;

 bad_exit:
  conf_eth_offset = 0;
  fprintf(stderr, "Warning: couldn't obtain the name of the nic operated by GAMMA. Assuming eth0.\n");
}
#endif

void apply_offset(void)
{
  int i;

  for(i=0; i<num_nodes; i++)
    nic[i] -= my_first_nic_index;
}



/**********
 * OUTPUT *
 **********/

void print_parsed_tables(FILE *f)
{
  int i,j;
  
  for (i=0; i < num_subnets; i++) 
    {
      fprintf(f, "subnet {\n");
      for (j=0; j < subnets[i].num_entries; j++) 
	{
	  fprintf(f, "\t%s\t%seth%d\n", subnets[i].entry_name[j], 
		 (strlen(subnets[i].entry_name[j]) >= 8)? "" : "\t", 
		 subnets[i].entry_nic[j]);
	}
      fprintf(f, "}\n\n");
    }
}

void print_parsed_flat_tables(FILE *f)
{
  int j;
  
  fprintf(f, "net {\n");
  for (j=0; j < subnets[0].num_entries; j++) 
    fprintf(f, "\t%s\n", subnets[0].entry_name[j]);
  fprintf(f, "}\n\n");
}

void print_nodes(FILE *f)
{
  int i;
  
  for (i=0; i < num_nodes; i++) 
    {
      fprintf(f, "%s\n", host_table[i]);
    }
  fprintf(f, "\n");
}

void print_multiple_tables(FILE *f)
{
  int i, k;
  
  for (i=0; i < num_nodes; i++) 
    {
      fprintf(f, "%s\t%s", host_table[i], 
	     (strlen(host_table[i]) >= 8)? "" : "\t");
      for (k=0; k < routes_num[i]; k++)
	fprintf(f, "(%d,%d) ", routes[i][k].local,  routes[i][k].remote);
      fprintf(f, "\n");
    }
  fprintf(f, "\n");
}

#ifdef NEW_GAMMA_CONF

void print_final_fnn_tables(FILE *f)
{
  int i, k;
  
  for (i=0; i < num_nodes; i++) 
    {
      fprintf(f, "%s\t%s", host_table[i], 
	     (strlen(host_table[i]) >= 8)? "" : "\t");
      for (k=0; k<6; k++)
	fprintf(f, "%2.2X%s", eth[i].addr[k], (k < 5)? ":" : "\t");
      fprintf(f, "%d\n", nic[i]);
    }
  fflush(f);
}

void print_final_flat_tables(FILE *f)
{
  int i, k;
  
  for (i=0; i < num_nodes; i++) 
    {
      fprintf(f, "%s\t%s", host_table[i], 
	     (strlen(host_table[i]) >= 8)? "" : "\t");
      for (k=0; k<6; k++)
	fprintf(f, "%2.2X%s", eth[i].addr[k], (k < 5)? ":" : "\n");
    }
  fflush(f);  
}

#else

void print_final_fnn_tables(FILE *f)
{
  int i, k;
  
  for (i=0; i < num_nodes; i++) 
    {
      fprintf(f, "%s\t%s", host_table[i], 
	     (strlen(host_table[i]) >= 8)? "" : "\t");
      for (k=0; k<6; k++)
	fprintf(f, "0x%2.2X%s", eth[i].addr[k], (k < 5)? " " : "\t");
      fprintf(f, "%d\n", nic[i]);
    }
  fflush(f);
}

void print_final_flat_tables(FILE *f)
{
  int i, k;
  
  for (i=0; i < num_nodes; i++) 
    {
      fprintf(f, "%s\t%s", host_table[i], 
	     (strlen(host_table[i]) >= 8)? "" : "\t");
      for (k=0; k<6; k++)
	fprintf(f, "0x%2.2X%s", eth[i].addr[k], (k < 5)? " " : "\n");
    }
  fflush(f);
}

#endif



/********
 * MAIN *
 ********/

int gamma_fnn_conf()
{
  int ret = 0;
  char hostname[MAX_LEN];
  FILE *f_in;
  FILE *f_out;

  
  if (conf_hostname == NULL)
    gethostname(hostname, MAX_LEN);
  else
    strncpy(hostname, conf_hostname, MAX_LEN);

  srand(time(NULL));

  subnets = malloc(MAX_NUM_NODES * sizeof(struct subnet));
  num_nodes = num_subnets = my_num_nics = my_index = 0;

  /* parse input */
  if ((f_in = fopen(conf_file_input, "r")) == NULL)
    {
      fprintf(stderr, "%s: error opening file \"%s\"\n", 
	      strerror(errno), conf_file_input);
      ret = 1;
      goto exit;
    }
	
  if (parse_file(f_in) < 0) 
    {
      fclose(f_in);
      ret = 2;
      goto exit;
    }

  fclose(f_in);

  /* check wheter flat or fnn GAMMA conf file must be built */
#if 0
  if (num_subnets == 1)
    {
      if (conf_very_verbose)
	{
	  fprintf(stdout, "Parsed table:\n");
	  print_parsed_flat_tables(stdout);
	}
      if (conf_very_verbose && parsed_flat_syntax)
	  fprintf(stdout, "Flat GAMMA configuration found.\n");
      if (!parsed_flat_syntax)
	fprintf(stderr, "Warning: Found only one net, but fnn syntax was used.\n"
		"Building flat GAMMA conf.\n");

      if (conf_eth_offset == -1)
	get_offset();
      if (conf_very_verbose)
	fprintf(stdout, "eth%d is the nic operated by GAMMA.\n", conf_eth_offset);

      my_first_nic_index = conf_eth_offset;

      compute_flat_nodes();
    }
  else
#endif
    {
      if (conf_very_verbose)
	{
	  fprintf(stdout, "Parsed table:\n");
	  print_parsed_tables(stdout);
	}

      /* get node list */
      if (conf_very_verbose)
	{
	  fprintf(stdout, "fnn GAMMA configuration found.\n"
		  "Computing nodes list... \n");
	  fflush(stdout);
	}
      if (compute_fnn_nodes(hostname) < 0)
	{
	  ret = 3;
	  goto exit;
	}
      if (conf_very_verbose)
	{
	  fprintf(stdout, "I'm number %d. Index of my first nic %d. Nodes:\n",
		  my_index, my_first_nic_index);
	  print_nodes(stdout);
	}

      /* compute all possible routes */
      fill_tables(hostname);
      if (conf_very_verbose)
	{
	  fprintf(stdout, "Multiple routes table:\n");
	  print_multiple_tables(stdout);
	}

      /* resolve multiple routes */
      if (conf_very_verbose)
	{
	  fprintf(stdout, "Resolving multiple routes... \n");
	  fflush(stdout);
	}

      (conf_random)?
	resolve_routes_random(): 
	resolve_routes_static();
    }

  /* get MAC addresses */
  if (conf_no_eth_res)
    {
      if (conf_very_verbose)
        fprintf(stdout, "Skipping resolving of ethernet addresses...  \n");
    }
  else
    {
      if (conf_very_verbose)
	fprintf(stdout, "Resolving ethernet addresses... \n");
      resolve_ethernet_addresses();
    }

  /* apply the index of the first nic */
  apply_offset();

  /* output */
  if (!conf_try) 
    {
      if ((f_out = fopen(conf_file_output, "w")) == NULL)
	{
	  fprintf(stderr, "%s: error opening file \"%s\"\n", 
		  strerror(errno), conf_file_output);
	  ret = 5;
	  goto exit;
	}
      if (num_subnets == 1)
	print_final_flat_tables(f_out);
      else
	print_final_fnn_tables(f_out);
      fclose(f_out);
    }

  if (conf_verbose)
    {
      if (num_subnets == 1)
	{
	  fprintf(stdout, "\nGAMMA configuration file:\n");
	  print_final_flat_tables(stdout);
	}
      else
	{
	  fprintf(stdout, "\nGAMMA configuration file for node %s is:\n", 
		    host_table[my_index]); 
	  print_final_fnn_tables(stdout);
	}
    }

 exit:
  free(subnets);
  return ret;
}

void print_help(FILE *f)
{
  fprintf(f, "\n"
	  "Usage: %s [OPTIONS]...\n\n"
	  "-t\t\ttest only. no output is written\n"
	  "-f FILE\t\tread given file instead of /etc/gamma.global\n"
	  "-o FILE\t\twrite to given file instead of /etc/gamma.conf\n"
	  "-e N\t\tWhen FLAT GAMMA is built, force N as the nic operated\n"
	  "\t\tby GAMMA instead of guessing it via dmesg\n"
	  "-u\t\tDo not get MAC addresses via ssh, overriding the default.\n"
	  "-n HOSTNAME\tuse given hostname and not the machine name\n"
	  "-v\t\tverbose\n"
	  "-V\t\tvery verbose\n"
	  "-r\t\trandomize multiple routes\n"
	  "-h\t\tprint this help and exit\n\n",
	  conf_program_name);
}


void app_init(int argc, char *argv[])
{
  conf_program_name = argv[0];

  for(argc--,argv++; argc--; argv++)
    {
      if( (*argv)[0] == '-')
        {
          switch( (*argv)[1] )
            {
	    case 'e': // conf_eth_offset
	      conf_eth_offset = atoi(*++argv);
	      argc--;
	      break;
	    case 'u': // conf_no_eth_res
	      conf_no_eth_res = 1;
	      break;
            case 't': // conf_try
	      conf_try = 1;
              break;
            case 'f': // *conf_file_input
	      conf_file_input = *++argv;
	      argc--;
              break;
            case 'o': // *conf_file_output
	      conf_file_output = *++argv;
              argc--;
              break;
            case 'v': // conf_verbose
	      conf_verbose = 1;
              break;
            case 'V': // conf_very_verbose
	      conf_very_verbose = conf_verbose = 1;
              break;
            case 'n': // *conf_hostname
	      conf_hostname = *++argv;
              argc--;
              break;
            case 'r': // conf_random
	      conf_random = 1;
              break;
            case 'h':
              print_help(stdout);
              exit(0);
            default:
              goto die;
            }
        }
    }
  return;

 die:
  fprintf(stderr, "Bad usage. try %s -h to see available arguments.n\n", 
	  conf_program_name);
  exit(1);
}


int main(int argc, char *argv[]) 
{
  app_init(argc, argv);
  return gamma_fnn_conf();
}
